"""
mkinit netharn.data
"""
# flake8: noqa

__DYNAMIC__ = False
if __DYNAMIC__:
    from mkinit import dynamic_init
    exec(dynamic_init(__name__))
else:
    # <AUTOGEN_INIT>
    from netharn.data import base
    from netharn.data import batch_samplers
    from netharn.data import coco_api
    from netharn.data import collate
    from netharn.data import mnist
    from netharn.data import toydata
    from netharn.data import transforms
    from netharn.data import voc

    from netharn.data.base import (DataMixin,)
    from netharn.data.batch_samplers import (MatchingSamplerPK,)
    from netharn.data.coco_api import (CocoDataset,)
    from netharn.data.collate import (CollateException, default_collate,
                                      list_collate, numpy_type_map,
                                      padded_collate,)
    from netharn.data.mnist import (MNIST,)
    from netharn.data.toydata import (ToyData1d, ToyData2d,)
    from netharn.data.voc import (VOCDataset,)

    __all__ = ['CocoDataset', 'CollateException', 'DataMixin', 'MNIST',
               'MatchingSamplerPK', 'ToyData1d', 'ToyData2d', 'VOCDataset', 'base',
               'batch_samplers', 'coco_api', 'collate', 'default_collate',
               'list_collate', 'mnist', 'numpy_type_map', 'padded_collate',
               'toydata', 'transforms', 'voc']
