# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import ubelt as ub

CacheStamp = ub.CacheStamp


if __name__ == '__main__':
    """
    CommandLine:
        python -m netharn.util.util_cachestamp all
    """
    import xdoctest
    xdoctest.doctest_module(__file__)
